// Title:   DeformerTool.cpp
// Created: Thu Sep 25 15:07:33 2003
// Authors: Richard Keiser, Mark Pauly
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <qwhatsthis.h>
#include <qapplication.h>
#include <qradiobutton.h>
#include <qspinbox.h>
#include <qpainter.h>
#include <qcheckbox.h>
#include <qmessagebox.h>
#include <qlabel.h>

#include "../../../../Core/Scene/src/Object.h"
#include "../../../../Core/Scene/src/Scene.h"
#include "../../../../Core/Configuration/src/Configuration.h"
#include "../../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../../Core/DataStructures/src/AdjacencyGraph.h"
#include "../../../../Core/CoreTools/src/CoreTools.h"
#include "../../../../Core/CoreResourceTools/src/CoreResourceTools.h"
#include "../../../../Core/Interfaces/src/ToolInterface.h"
#include "../../../../Core/Interfaces/src/FrameBufferInterface.h"
#include "../../../../Core/Interfaces/src/RendererInterface.h"
#include "../../../../Utilities/src/Common.h"
#include "../../../../Utilities/src/MyMathTools.h"
#include "../../../../Utilities/UserMessages/src/StatusBar.h"
#include "../../../../Utilities/src/TrackBall.h"
#include "../../../../Core/ToolManager/src/ToolManager.h"

#include "../UI/DeformerConfigurationWidget.h" 
#include "DeformerTool.h"
#include "DeformSurfel.h"

#define PIE_SIZE 20

// ******************
// 'What's this' text
// ******************

static const char *whatsThisText = "<img source=\"deformer\"> "
"Free form deformation tool:"
"<ul><li>left mouse button: perform deformation</li>"
"<li>middle mouse button: change deformation </li></ul>";

DeformerTool::DeformerTool() {

	standardCursor = new QCursor (CrossCursor);
	transformCursor = new QCursor (SizeAllCursor);
	toolCursor     = standardCursor;	

	QPixmap normalMoveIcon ("Resources/DeformerToolMoveIcon.png"),
		    disabledMoveIcon ("Resources/DeformerToolMoveIcon_disabled.png"),
			normalRotateIcon ("Resources/DeformerToolRotationIcon.png"),
			disabledRotateIcon ("Resources/DeformerToolRotationIcon_disabled.png");
	
	moveIconSet.setPixmap (normalMoveIcon, QIconSet::Automatic, QIconSet::Normal);
	moveIconSet.setPixmap (disabledMoveIcon, QIconSet::Automatic, QIconSet::Disabled);

	rotateIconSet.setPixmap (normalRotateIcon, QIconSet::Automatic, QIconSet::Normal);
	rotateIconSet.setPixmap (disabledRotateIcon, QIconSet::Automatic, QIconSet::Disabled);

	usePreviewRenderer = true;	
	zombiePositions = NULL;
	zombieColors = NULL;
	numZombies = 0;
	
	MtrUnity4x4f (axisScaleTranslationMatrix);
	MtrUnity4x4f (axisRotationMatrix);

	frameShape = QPointArray (4);	
	rotationPieShape = QPointArray (PIE_SIZE);
	rotationAngle = M_PI / 2.0f;

	orthoMapper			= NULL;
	zeroRegion			= NULL;
	zeroRegionPositions = NULL;
	oneRegion			= NULL;
}


DeformerTool::~DeformerTool() {

	delete standardCursor;
	delete transformCursor;
	delete configurationWidget;
	if (orthoMapper) {
		delete orthoMapper;	
		orthoMapper = NULL;
	}
	if (zeroRegion) {
		delete zeroRegion;
	}
	if (zeroRegionPositions) {
		delete[] zeroRegionPositions;
	}
	if (oneRegion) {
		delete oneRegion;
	}
}

// **************
// public methods
// **************

QToolButton *DeformerTool::addButtonToToolBar (QToolBar *toolBar) {


	toolButton = new QToolButton (moveIconSet, "Deformer Tool (<b>D</b>)", "Perform free form deformations",
	                              0, 0, toolBar, "Deformer Tool");
	toolButton->setToggleButton (true);
	QWhatsThis::add (toolButton, whatsThisText);
	QMimeSourceFactory::defaultFactory()->setPixmap ("deformer", moveIconSet.pixmap (QIconSet::Automatic, QIconSet::Normal));

	moveID   = deformerMenu.insertItem (moveIconSet.pixmap (QIconSet::Automatic, QIconSet::Normal), this, SLOT (handleToolSelected(int)));
	rotateID = deformerMenu.insertItem (rotateIconSet.pixmap (QIconSet::Automatic, QIconSet::Normal), this, SLOT (handleToolSelected(int)));
	
	currentTool = moveID;

	toolButton->setPopup (&deformerMenu);
	toolButton->setEnabled (this->checkToolAvailability());

	configurationWidget = new DeformerConfigurationWidget;
	configurationWidget->previewRendererCheckBox->setChecked (usePreviewRenderer);

	
	return toolButton;
}


QCursor *DeformerTool::getToolCursor() const {
	return toolCursor;
}


QToolButton* DeformerTool::getToolButton () {

	return toolButton;
}



QWidget *DeformerTool::getToolConfigurationWidget() const {
	return configurationWidget;
}


bool DeformerTool::isAvailable() const {
	return this->checkToolAvailability();
} 



void DeformerTool::handleMouseMoveEvent (QMouseEvent *me) {

	ToolInterface::handleMouseMoveEvent(me);

	if (!isGlobalTransformationActive()) {

		if (currentTool == moveID) {
			
			float mouseMoveDistance = me->y() - lastMouseDragPosition.y();
			
			if (me->state() == Qt::MidButton) {		
				orthoMapper->adjustCursorShape (OrthoMapper::GLOBAL_SIZE, (int)mouseMoveDistance);
			}
			else if (me->state() == (Qt::MidButton | Qt::ShiftButton)) {
				orthoMapper->adjustCursorShape (OrthoMapper::X_SIZE, (int)mouseMoveDistance);
			}	
			else if (me->state() == (Qt::MidButton | Qt::ControlButton)) {
				orthoMapper->adjustCursorShape (OrthoMapper::Y_SIZE, (int)mouseMoveDistance);
			}	
			else if (me->state() == (Qt::MidButton | Qt::AltButton)) {
				orthoMapper->adjustCursorShape (OrthoMapper::Z_SIZE, (int)mouseMoveDistance);
			}
			else if (me->state() == Qt::RightButton) {
				orthoMapper->adjustCursorShape (OrthoMapper::ORIENTATION, (int)mouseMoveDistance);
			}
			else {
				screenPosition[0] = me->x();
				screenPosition[1] = me->y();
			}
			
			if ( (configurationWidget->deformationContinuousRadioButton->isChecked()) && (me->state() & Qt::LeftButton) ) {
				deformMoveScale = 0.01f * (lastMouseDragPosition.y() - me->y());

				DeformParams deformParams = computeDeformParams(me);
		
				this->deformModel (deformParams);

				if (configurationWidget->dynamicResamplingCheckBox->isChecked() && configurationWidget->deformationContinuousRadioButton->isChecked()) {
					splitSurfels (2.0f);
				}

				lastMouseDragPosition = me->pos();	
				
			}
			else {
				
				currentPositionValid = orthoMapper->drawCursor (screenPosition, 0);
				
				if (currentPositionValid) {
					if (me->state() == Qt::LeftButton) {	
						if (configurationWidget->deformationDiscreteRadioButton->isChecked()) {

							deformMoveScale = orthoMapper->getZScale();
							deformMoveVector = orthoMapper->getToolNormalInObjectSpace();		

							DeformParams deformParams = computeDeformParams(me);
							
							this->deformModel (deformParams);
						}
					}		
				}
			}
		}
		else if (currentTool == rotateID) {

			screenPosition[0] = me->x();
			screenPosition[1] = me->y();
		
			currentPositionValid = orthoMapper->drawCursor (screenPosition, 0.0f);

			// scale rotation
			if (me->state() == (Qt::RightButton | Qt::ShiftButton)) {
				rotationAngle += 0.04f * (me->y() - lastMouseDragPosition.y());	
				this->updateRotationAxis();
			}
			// move frame in z-direction
			else if (me->state() == (Qt::RightButton | Qt::AltButton)) {
				axisScaleTranslationMatrix[14] += 2.0f * (me->y() - lastMouseDragPosition.y());			
				this->updateRotationAxis();
			}		
			// move frame in x- and y-direction
			else if (me->state() == (Qt::RightButton | Qt::ControlButton)) {				
				axisScaleTranslationMatrix[12] += me->x() - lastMouseDragPosition.x();
				axisScaleTranslationMatrix[13] += lastMouseDragPosition.y() - me->y();
				this->updateRotationAxis();				
			}				
			// rotate frame
			else if (me->state() == Qt::RightButton) {
				
				float normalizedMousePositions[4];
				float rotationVector[4];
				QSize viewPortSize;
				viewPortSize = Configuration::getInstance()->getViewPortSize();
				
				// normalize mouse positions into rectangle, [-1.0, 1.0] x [-1.0, 1.0]
				normalizedMousePositions[0] = (float)lastMouseDragPosition.x() / (float)viewPortSize.width()  * 2.0f - 1.0f;
				normalizedMousePositions[1] = (float)lastMouseDragPosition.y() / (float)viewPortSize.height() * 2.0f - 1.0f;
				normalizedMousePositions[2] = (float)me->x() / (float)viewPortSize.width()  * 2.0f - 1.0f;
				normalizedMousePositions[3] = (float)me->y() / (float)viewPortSize.height() * 2.0f - 1.0f;
				
				if (TrackBall::calculateRotationMatrix (normalizedMousePositions, rotationVector) == true) {
					
					MyDataTypes::TransformationMatrix16f userRotation,	finalRotation;
					
					MtrCreateRotation4x4fc (rotationVector[0] * PI_180, rotationVector[1], rotationVector[2], rotationVector[3], userRotation);
					MtrMult4x4f (userRotation, axisRotationMatrix, finalRotation);
					MtrCopy4x4f (finalRotation, axisRotationMatrix);				
				}
				this->updateRotationAxis();
			}			
			lastMouseDragPosition = me->pos();	
		}
	}	// if (!isGlobalTransformationActive())
	else {
		if (me->state() & (Qt::LeftButton | Qt::RightButton | Qt::MidButton)) {
			this->updateRotationAxis();
		}
	}
	
}


void DeformerTool::handleMouseReleaseEvent (QMouseEvent *) {

	if (!isGlobalTransformationActive()) {
	
		if (configurationWidget->dynamicResamplingCheckBox->isChecked() && configurationWidget->deformationContinuousRadioButton->isChecked()) {
			splitSurfels (2.0f);
			if (configurationWidget->automaticFairingCheckbox->isChecked()) {
				handleFairing();	
			}
			
			Scene::getInstance()->emitObjectModifiedSignal ();
			Scene::getInstance()->emitRenderFrameSignal ();
		}
		
	}
	
	RendererManager::getInstance()->setPreviewRendererEnabled (Configuration::getInstance()->isPreviewRendererEnabled(), true);	

}



void DeformerTool::handleMousePressEvent (QMouseEvent *me) {

	ToolInterface::handleMousePressEvent(me);
	lastMouseDragPosition = me->pos();

	if (!isGlobalTransformationActive()) {

		if (currentTool == moveID) {
			if (currentPositionValid) {

				if (me->button() == Qt::LeftButton && me->state() == Qt::NoButton) {

					if (configurationWidget->deformationContinuousRadioButton->isChecked()) {
					
						RendererManager::getInstance()->setPreviewRendererEnabled (usePreviewRenderer, true);
						deformMoveVector = orthoMapper->getZScale() * orthoMapper->getToolNormalInObjectSpace();		

					}
					else if (configurationWidget->deformationDiscreteRadioButton->isChecked()) {

						deformMoveScale = orthoMapper->getZScale() * 4.0f;
						deformMoveVector = orthoMapper->getToolNormalInObjectSpace();		

						DeformParams deformParams = computeDeformParams(me);
						this->deformModel (deformParams);
					}
				}
				else if (me->button() == Qt::MidButton) {			
					orthoMapper->initAdjustCursorShape();
					lastMouseDragPosition = me->pos();			
				}
				else if (me->button() == Qt::RightButton) {
					splitSurfels (2.0f);
				}
			}
		}
		else if (currentTool == rotateID) {

			if ((me->button() == Qt::LeftButton) || (me->button() == Qt::MidButton)) {
				
				if (currentPositionValid) {
					Transformer *transformer = Transformer::getSceneTransformer();
					RendererInterface *renderer = RendererManager::getInstance()->getMainRenderer();				
					
					Vector3D rotCenter = transformer->viewPortToScene (screenPosition, renderer->getDepth (screenPosition[0], screenPosition[1]));				
					
					axisScaleTranslationMatrix[12] = rotCenter[0];
					axisScaleTranslationMatrix[13] = rotCenter[1];
					axisScaleTranslationMatrix[14] = rotCenter[2];			
					
					if (me->button() == Qt::MidButton) {
						
						Vector3D rotAxis = orthoMapper->getToolNormalInObjectSpace();
						Vector3D rotAxisScene = transformer->objectToSceneVector (rotAxis);
						rotAxisScene.normalize();
						Vector3D axisOne, axisTwo;
						rotAxisScene.getTwoOrthogonals (axisOne, axisTwo);
						axisOne.normalize();
						axisTwo.normalize();
					
						if (me->state() == (Qt::ControlButton)) {
							axisRotationMatrix[4] = axisOne[0];
							axisRotationMatrix[5] = axisOne[1];
							axisRotationMatrix[6] = axisOne[2];

							axisRotationMatrix[8] = axisTwo[0];
							axisRotationMatrix[9] = axisTwo[1];
							axisRotationMatrix[10] = axisTwo[2];

							axisRotationMatrix[0] = rotAxisScene[0];
							axisRotationMatrix[1] = rotAxisScene[1];
							axisRotationMatrix[2] = rotAxisScene[2];

						}
						else if (me->state() == (Qt::ShiftButton)) {
							axisRotationMatrix[8] = axisOne[0];
							axisRotationMatrix[9] = axisOne[1];
							axisRotationMatrix[10] = axisOne[2];

							axisRotationMatrix[0] = axisTwo[0];
							axisRotationMatrix[1] = axisTwo[1];
							axisRotationMatrix[2] = axisTwo[2];

							axisRotationMatrix[4] = rotAxisScene[0];
							axisRotationMatrix[5] = rotAxisScene[1];
							axisRotationMatrix[6] = rotAxisScene[2];
						}
						else {
							axisRotationMatrix[0] = axisOne[0];
							axisRotationMatrix[1] = axisOne[1];
							axisRotationMatrix[2] = axisOne[2];

							axisRotationMatrix[4] = axisTwo[0];
							axisRotationMatrix[5] = axisTwo[1];
							axisRotationMatrix[6] = axisTwo[2];

							axisRotationMatrix[8] = rotAxisScene[0];
							axisRotationMatrix[9] = rotAxisScene[1];
							axisRotationMatrix[10] = rotAxisScene[2];
						}

						bIsSetRotationAxis = true;

					}
		
					this->updateRotationAxis();
					
				}
			}

		}
	}
	else {
		RendererManager::getInstance()->setPreviewRendererEnabled (usePreviewRenderer, true);
	}

}


void DeformerTool::handleKeyPressEvent (QKeyEvent *ke) {
	
	ToolInterface::handleKeyPressEvent(ke);
	
	if (ke->state() == ControlButton && ke->key() == Qt::Key_Z)	{
		undoDeformation();
	}
	else if (ke->key() == Qt::Key_F) {
		handleFairing();
	}
	else if (ke->key() == Qt::Key_I) {
		handleInterpolation();
	}
	else if (ke->key() == Qt::Key_P) {
		configurationWidget->previewRendererCheckBox->setChecked(!usePreviewRenderer);
	}
	else if (ke->key() == Qt::Key_R) {
		configurationWidget->dynamicResamplingCheckBox->setChecked(!useDynamicResampling);
	}
	else if (ke->key() == Qt::Key_Z) {
		handleShowZombiesChanged();
	}
}

void DeformerTool::handleKeyReleaseEvent (QKeyEvent *ke) {
	ke->ignore();
}

void DeformerTool::selectTool() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);

	mainRenderer = RendererManager::getInstance()->getMainRenderer();
	orthoMapper = new OrthoMapper (QSize (100, 100), OrthoMapper::PLAIN);

	this->connect (configurationWidget->previewRendererCheckBox, SIGNAL (toggled (bool)),
		           this, SLOT (handlePreviewRendererToggled (bool)));

	this->connect (configurationWidget->fairingPushButton, SIGNAL (clicked()),
		           this, SLOT (handleFairing ()));
		
	this->connect (Scene::getInstance(), SIGNAL (sceneMoved (const MyDataTypes::TransformationMatrix16f)),
					this, SLOT (handleSceneMoved ()));

	this->connect (Configuration::getInstance(), SIGNAL (viewPortSizeChanged (const QSize)),
					this, SLOT (handleViewPortSizeChanged ()));

	this->connect (Scene::getInstance(), SIGNAL (activeObjectChanged()),
					this, SLOT (handleActiveModelChanged()));

	this->connect( configurationWidget->ShowRotationAxesCheckBox, SIGNAL(stateChanged(int)), 
					this, SLOT(showRotationAxes()) );

	this->connect (configurationWidget->dynamicResamplingCheckBox, SIGNAL (toggled (bool)),
					this, SLOT (handleDynamicResamplingToggled (bool)));

	this->connect (configurationWidget->showZombiesCheckBox, SIGNAL (stateChanged(int)),
					this, SLOT (handleShowZombiesChanged()));

	this->connect (configurationWidget->interpolateColorsPushButton, SIGNAL (clicked()),
					this, SLOT (handleInterpolation()));


	// turn mouse tracking on
	RendererManager::getInstance()->getMainRenderer()->getRendererWidget()->setMouseTracking (true);

	zeroRegionPositions = 0;
	oneRegionPositions  = 0;
	useDynamicResampling = configurationWidget->dynamicResamplingCheckBox->isChecked();
	
	if (!this->initDeformation()) {
		ToolManager::getInstance()->switchToNavigationTool();
	}
	else {
		this->renderRotationAxis();
		setGlobalTransformationActive(false);
	}
	
	QApplication::restoreOverrideCursor ();
}


void DeformerTool::deselectTool() {

	// turn mouse tracking off
	RendererManager::getInstance()->getMainRenderer()->getRendererWidget()->setMouseTracking (false);
	
	if (orthoMapper) {
		delete orthoMapper;	
		orthoMapper = NULL;
	}
	if (zeroRegion) {
		delete zeroRegion;
		zeroRegion = NULL;
	}
	if (zeroRegionPositions) {
		delete[] zeroRegionPositions;
		zeroRegionPositions = NULL;
	}
	if (oneRegion) {
		delete oneRegion;
		oneRegion = NULL;
	}

	if (zombiePositions) {
		delete[] zombiePositions;
		zombiePositions = NULL;
		numZombies = 0;
		delete[] zombieColors;
		zombieColors = NULL;
	}

	this->disconnect (configurationWidget->previewRendererCheckBox, SIGNAL (toggled (bool)), this, SLOT (handlePreviewRendererToggled (bool)));
	this->disconnect (configurationWidget->fairingPushButton, SIGNAL (clicked()), this, SLOT (handleFairing ()));
	this->disconnect (Scene::getInstance(), SIGNAL (sceneMoved (const MyDataTypes::TransformationMatrix16f)), this, 0);
	this->disconnect (Configuration::getInstance(), SIGNAL (viewPortSizeChanged (const QSize)), this, 0);
	this->disconnect (Scene::getInstance(), SIGNAL (activeObjectChanged()), this, 0);
	this->disconnect (configurationWidget->ShowRotationAxesCheckBox, SIGNAL(stateChanged(int)), this, SLOT(showRotationAxes()) );
	this->disconnect (configurationWidget->dynamicResamplingCheckBox, SIGNAL (toggled (bool)), this, SLOT (handleDynamicResamplingToggled (bool)));
	this->disconnect (configurationWidget->showZombiesCheckBox, SIGNAL (stateChanged(int)), this, SLOT (handleShowZombiesChanged()));
	this->disconnect (configurationWidget->interpolateColorsPushButton, SIGNAL (clicked()), this, SLOT (handleInterpolation()));

}


// ************
// public slots
// ************

void DeformerTool::handleRendererWidgetRepainted () {

	orthoMapper->reset();
	if (!isGlobalTransformationActive()) {
		currentPositionValid = orthoMapper->drawCursor (screenPosition, 0);
	}
	this->renderRotationAxis();
}



void DeformerTool::handleFairing () {


	bool						isNum;
	unsigned int				nofNeighbors, iterations, numAllSurfels;
	float						damping, timeStep;		
	unsigned int				i, j, k;
	NeighbourHood				*searchRegion;
	Vector3D					*regionPositions;
	float						dist, maxDist;
	SurfelInterface				**neighbors,
								**tmpNeighbors,
								**regionSurfels,
								*surfel;
	Vector3D					pos, displacement, direction;
	SurfelCollection			*surfels;
	QVector<SurfelInterface>	deformRegionSurfels;
	Scene						*scene;
		

	QApplication::setOverrideCursor (Qt::waitCursor);	

	scene = Scene::getInstance();
	CoreTools::getInstance()->getSelectionTool()->getAllNotSelected (deformRegionSurfels, SurfelInterface::SELECTED1);
	numDeformSurfels = deformRegionSurfels.count();
	surfels			 = scene->getActiveObject()->getSurfelCollection();
	numAllSurfels	 = surfels->getNofSurfels();

	Vector3D* forces		= new Vector3D	[numDeformSurfels];
	regionPositions			= new Vector3D	[numAllSurfels];
	regionSurfels			= new SurfelInterface* [numAllSurfels];
	
	timeStep     = configurationWidget->fairingTimestepLineEdit->text().toFloat (&isNum);
	iterations   = configurationWidget->fairingIterationsSpinBox->value();
	damping     = configurationWidget->fairingDampingLineEdit->text().toFloat (&isNum);
	nofNeighbors = configurationWidget->interpolationNeighborsSpinBox->value();

	neighbors = new SurfelInterface*[numDeformSurfels*nofNeighbors];
	tmpNeighbors = neighbors;

	// create search structures for nearest point queries
	for (i = 0, surfel = surfels->getFirstSurfel(); surfel; surfel = surfels->getNextSurfel(), i++) {
		regionPositions[i] = surfel->getPosition();
		regionSurfels[i]   = surfel;
	}

	searchRegion = new NeighbourHood (regionPositions, numAllSurfels);
	searchRegion->setNofNeighbours (nofNeighbors+1);

	// cache neighbors
	for (i = 0; i < numDeformSurfels; i++) {				
		searchRegion->setSourcePoint (deformRegionSurfels.at(i)->getPosition());
		for (j=nofNeighbors; j>0; j--, tmpNeighbors++) {
			*tmpNeighbors = regionSurfels[searchRegion->getNeighbourPositionIndex(j)];
		}
	}
	
	for (k = 0; k < iterations; k++) {

		tmpNeighbors = neighbors;

		// compute forces on particles
		for (i = 0; i < numDeformSurfels; i++) {			
				
			forces[i].makeZero();
			pos = deformRegionSurfels.at(i)->getPosition();
	
			maxDist = Vector3D::distance (pos, (*tmpNeighbors)->getPosition());
			tmpNeighbors++;

			for (j = 1; j < nofNeighbors; j++, tmpNeighbors++) {						
				
				direction = pos - (*tmpNeighbors)->getPosition();
				dist = direction.normalize();
				
				forces[i] += (maxDist - dist) * direction;				
			}
		}

		// display particles according to the forces
		for (i = 0; i < numDeformSurfels; i++) {
			Vector3D normal = deformRegionSurfels.at(i)->getNormal();
			displacement = Vector3D::projectVectorIntoPlane (forces[i], normal);
			scene->setSurfelPosition (deformRegionSurfels.at(i), 
									  deformRegionSurfels.at(i)->getPosition() + timeStep * displacement, false);
		}
		
		timeStep *= damping;
	}

	Scene::getInstance()->emitObjectModifiedSignal ();
	Scene::getInstance()->emitRenderFrameSignal ();

	delete[] forces;
	delete[] regionPositions;
	delete   searchRegion;
	delete[] neighbors;
	delete[] regionSurfels;

	QApplication::restoreOverrideCursor ();
}



// ***************
// private methods
// ***************
void DeformerTool::setGlobalTransformationActive(bool isOn) {

	ToolInterface::setGlobalTransformationActive(isOn);
	if (!isOn) {
		toolCursor = standardCursor;
	}
	else {
		toolCursor = transformCursor;
	}

	RendererManager *rendererManager;
	QWidget         *rendererWidget;

	rendererManager = RendererManager::getInstance();
	if (rendererManager->isPreviewRendererEnabled()) {
		rendererWidget = rendererManager->getPreviewRenderer()->getRendererWidget();
		rendererWidget->setCursor (*toolCursor);
	}
	else {
		rendererWidget = rendererManager->getMainRenderer()->getRendererWidget();
		rendererWidget->setCursor (*toolCursor);
	}

	Scene::getInstance()->emitRenderFrameSignal ();
}


bool DeformerTool::initDeformation () {	

	Scene* scene = Scene::getInstance();

	if (scene->getActiveObject() != 0) {

		QVector<SurfelInterface> *zeroRegionSurfels,     // vector of surfels returned by selection tool
		                         *oneRegionSurfels,
							     *scaleRegionSurfels;
		QVector<SurfelInterface> deformRegionSurfels;
		Vector3D                 *deformRegionPositions,
								 *oneRegionPositions,
								 *scaleRegionPositions;
		NeighbourHood            *scaleRegion;

		unsigned int             i;
		StatusBar                *statusBar;
		

		statusBar = StatusBar::getInstance();		
		statusBar->showMessage ("Initializing deformation...", -1);
		statusBar->resetProgress();

		bIsSetRotationAxis = false;
		currentPositionValid = false;		
		// set to NONE so that deformation scales get computed when deformModel is first called
		currentBlendingFunction = DeformSurfel::NONE;
								
		objectSurfels = scene->getActiveObject()->getSurfelCollection();
						
		// zero region is defined by surfels that are selected with flag SELECTED1 (red region)		
		zeroRegionSurfels  = CoreTools::getInstance()->getSelectionTool()->getSelection (SurfelInterface::SELECTED1);

		// one region is defined by surfels that are selected with flag SELECTED2 (green region)		
		oneRegionSurfels   = CoreTools::getInstance()->getSelectionTool()->getSelection (SurfelInterface::SELECTED2);
		
		// deform region is defined by all surfels that are NOT selected with flag SELECTED1 (red region)		
		CoreTools::getInstance()->getSelectionTool()->getAllNotSelected (deformRegionSurfels, SurfelInterface::SELECTED1);
		numDeformSurfels = deformRegionSurfels.count();

		if (deformRegionSurfels.count() == 0 || zeroRegionSurfels->count() == 0 || oneRegionSurfels->count() == 0) {
			if (deformRegionSurfels.count() == 0) {
				statusBar->showMessage ("No deform region! Could not initialize the deformation tool!", -1);
			}
			else if (zeroRegionSurfels->count() == 0) {
				statusBar->showMessage ("No zero region! Could not initialize the deformation tool!", -1);
			}
			else if (oneRegionSurfels->count() == 0) {
				statusBar->showMessage ("No one region! Could not initialize the deformation tool!", -1);
			}
			return false;
		}
		
		statusBar->setProgress (10);
		
		deformRegionPositions = new Vector3D [deformRegionSurfels.count()];
		zeroRegionPositions   = new Vector3D [zeroRegionSurfels->count()];
		oneRegionPositions    = new Vector3D [oneRegionSurfels->count()];
		

		// create search structures for nearest point queries
		for (i = 0; i < deformRegionSurfels.count(); i++) {
			deformRegionPositions[i] = deformRegionSurfels.at(i)->getPosition();
		}

		statusBar->setProgress (15);

		for (i = 0; i < zeroRegionSurfels->count(); i++) {
			zeroRegionPositions[i] = zeroRegionSurfels->at(i)->getPosition();
		}
		zeroRegion = new NeighbourHood (zeroRegionPositions, zeroRegionSurfels->count());

		statusBar->setProgress (20);
		
		for (i = 0; i < oneRegionSurfels->count(); i++) {
			oneRegionPositions[i] = oneRegionSurfels->at(i)->getPosition();
		}						
		oneRegion  = new NeighbourHood (oneRegionPositions,  oneRegionSurfels->count());
		
		statusBar->setProgress (25);

			
		// create deformation scale factors for sample center and tangent axes		
		numScaleNeighbors = configurationWidget->numScaleNeighborsSpinBox->value();

		// limit the number of scale neighbors to min region counts
		if (zeroRegionSurfels->count() < numScaleNeighbors) {
			numScaleNeighbors = zeroRegionSurfels->count();
		}
		if (oneRegionSurfels->count() < numScaleNeighbors) {
			numScaleNeighbors = oneRegionSurfels->count();
		}

		zeroRegion->setNofNeighbours (numScaleNeighbors);
		oneRegion->setNofNeighbours (numScaleNeighbors);


		// build scale grid
		scaleGrid = new ScaleGrid (deformRegionPositions, deformRegionSurfels.count(),
			__min(35, deformRegionSurfels.count()), zeroRegion, oneRegion);
		
		deformSurfels = new QArray<DeformSurfel> (numDeformSurfels);
		
		for (i = 0; i < numDeformSurfels; i++) {			
			deformSurfels->at(i).initialize (deformRegionSurfels.at(i));
			deformSurfels->at(i).compNormDists (scaleGrid);

			if (i % (numDeformSurfels / 100) == 0) {
				statusBar->setProgress ((int)((float) i / (float) numDeformSurfels * 50.0f) + 25);
			}			
		}


		// scale deformation scales (additional Scaling factor determined by surfels that are 
		// selected with flag SELECTED3 (green region)
		if (CoreTools::getInstance()->getSelectionTool()->getSelectionSize(SurfelInterface::SELECTED3) > 0) {
			
			scaleRegionSurfels = CoreTools::getInstance()->getSelectionTool()->getSelection (SurfelInterface::SELECTED3);
			
			scaleRegionPositions = new Vector3D [scaleRegionSurfels->count()];
			for (i = 0; i < scaleRegionSurfels->count(); i++) {
				scaleRegionPositions[i] = scaleRegionSurfels->at(i)->getPosition();
			}
			scaleRegion = new NeighbourHood (scaleRegionPositions,  scaleRegionSurfels->count());
			scaleRegion->setNofNeighbours (numScaleNeighbors);
			
			float maxScaleDistance = 0.0f;
			for (i = 0; i < numDeformSurfels; i++) {
				
				deformSurfels->at(i).computeScale (scaleRegion, maxScaleDistance);
				
				if (i % (numDeformSurfels / 100) == 0) {
					statusBar->setProgress ((int)((float) i / (float) numDeformSurfels * 20.0f) + 75);
				}
			}
			
			for (i = 0; i < numDeformSurfels; i++) {
				deformSurfels->at(i).rescaleScale (maxScaleDistance);
			}

			delete[] scaleRegionPositions;
			delete scaleRegion;		
		}

		this->handleDynamicResamplingToggled(useDynamicResampling);

		delete[] deformRegionPositions;
		delete[] oneRegionPositions;

		statusBar->setProgress (100);
		statusBar->showMessage ("", -1);

	}

	return true;
}



void DeformerTool::updateRotationAxis () {
	
	renderRotationAxis();

	Vector3D pointScene;
	Vector2D pointViewSpace;

	MtrMult4x4f (axisScaleTranslationMatrix, axisRotationMatrix, axisTransformationMatrix);

	Transformer *transformer = Transformer::getSceneTransformer();

	// center
	MtrMultVector3DHom (axisTransformationMatrix, Vector3D (0, 0, 0), &pointScene);
	pointViewSpace = transformer->sceneToViewPort (pointScene);
	frameShape.setPoint (0, (int)pointViewSpace[0], (int)pointViewSpace[1] );

	// x-Axis
	MtrMultVector3DHom (axisTransformationMatrix, Vector3D (100, 0, 0), &pointScene);
	pointViewSpace = transformer->sceneToViewPort (pointScene);
	frameShape.setPoint (1, (int)pointViewSpace[0], (int)pointViewSpace[1] );

	// y-Axis
	MtrMultVector3DHom (axisTransformationMatrix, Vector3D (0, 100, 0), &pointScene);
	pointViewSpace = transformer->sceneToViewPort (pointScene);
	frameShape.setPoint (2, (int)pointViewSpace[0], (int)pointViewSpace[1] );

	// z-Axis
	MtrMultVector3DHom (axisTransformationMatrix, Vector3D (0, 0, 100), &pointScene);
	pointViewSpace = transformer->sceneToViewPort (pointScene);
	frameShape.setPoint (3, (int)pointViewSpace[0], (int)pointViewSpace[1] );

	// arc
	for (int i = 1; i < PIE_SIZE; i++) {
		MtrMultVector3DHom (axisTransformationMatrix, Vector3D (50.0f * cos ((i-1) * rotationAngle / (PIE_SIZE - 2)), 
			50.0f * sin ((i-1) * rotationAngle / (PIE_SIZE - 2)), 30), &pointScene);
		pointViewSpace = transformer->sceneToViewPort (pointScene);
		rotationPieShape.setPoint (i, (int)pointViewSpace[0], (int)pointViewSpace[1] );
	}
	MtrMultVector3DHom (axisTransformationMatrix, Vector3D (0, 0, 30), &pointScene);
	pointViewSpace = transformer->sceneToViewPort (pointScene);
	rotationPieShape.setPoint (0, (int)pointViewSpace[0], (int)pointViewSpace[1] );

	renderRotationAxis();

}



void DeformerTool::renderRotationAxis() {

	if (bIsSetRotationAxis && configurationWidget->ShowRotationAxesCheckBox->isChecked()) {

		QPainter painter;	
		
		painter.begin (RendererManager::getInstance()->getMainRenderer()->getRendererWidget());	
		painter.setPen (QPen (Qt::white, 1, Qt::SolidLine));
		painter.setFont (QFont ("", 15));		

		// XOR mode
		painter.setRasterOp (Qt::XorROP);
		painter.setBrush (QBrush (QColor (200, 200, 200)));		

		painter.drawLine (frameShape.at(0), frameShape.at(1));
		painter.drawLine (frameShape.at(0), frameShape.at(2));
		painter.drawLine (frameShape.at(0), frameShape.at(3));
		
		painter.drawText (frameShape.at(3), "Z");

		painter.drawPolygon (rotationPieShape);

		painter.end();
	}
	
}


void DeformerTool::deformModel (DeformParams &deformParams) {

	unsigned int i;

	// checks if deformation scales are up-to-date (blending function could have been changed by user)
	// and recomputes if necessary
	computeDeformationScales();

	for (i = 0; i < numZombies; i++) {
		zombiePositions[i] = DeformSurfel::deformPoint(zombiePositions[i], deformSurfels->at(i).getScalePosition(), deformParams);
	}
	
	for (i = 0; i < numDeformSurfels; i++) {						
		deformSurfels->at(i).deform (deformParams);
	}

	// split surfels
	if (configurationWidget->dynamicResamplingCheckBox->isChecked() && !configurationWidget->deformationContinuousRadioButton->isChecked()) {
		splitSurfels (2.0f);
	}

	Scene::getInstance()->emitObjectModifiedSignal (true);
	Scene::getInstance()->emitRenderFrameSignal ();
}

DeformParams DeformerTool::computeDeformParams(QMouseEvent *me) {
	
	bool isNumber;
	DeformParams deformParams;

	deformParams.doTranslation = configurationWidget->deformationTranslationCheckBox->isChecked();
	deformParams.doRotation = configurationWidget->deformationRotationCheckBox->isChecked();						

	deformParams.translation = deformMoveScale * deformMoveVector;			
		
	deformParams.angle = deformMoveScale * configurationWidget->deformationAngleLineEdit->text().toFloat(&isNumber);

	if (deformParams.doRotation) {

		Transformer *transformer = Transformer::getSceneTransformer();
		Vector3D rotationCenterScene, rotationAxisScene;

		MtrMultVector3DHom (axisTransformationMatrix, Vector3D (0, 0, 0), &rotationCenterScene);
		deformParams.rotationCenter = transformer->sceneToObject (rotationCenterScene);

		MtrMultVector3DHom (axisTransformationMatrix, Vector3D (0, 0, 1), &rotationAxisScene);
		rotationAxisScene -= rotationCenterScene;
		deformParams.rotationAxis = transformer->sceneToObjectVector (rotationAxisScene);
		deformParams.rotationAxis.normalize();
		deformParams.rotationAxis.getTwoOrthogonals (deformParams.rotationTangent1, deformParams.rotationTangent2);
		deformParams.rotationTangent1.normalize();
		deformParams.rotationTangent2.normalize();
	}

	return deformParams;
}


void DeformerTool::computeDeformationScales () {

	// compute deformation scales if a new blending function has been chosen

	DeformSurfel::BlendingFunction newBlendingFunction = DeformSurfel::NONE;
	unsigned int i;

	if (configurationWidget->functionSmoothRadioButton->isChecked()) {
		newBlendingFunction = DeformSurfel::SMOOTH;
	}
	else if (configurationWidget->functionSmooth2RadioButton->isChecked()) {
		newBlendingFunction = DeformSurfel::SMOOTH2;
	}
	else if (configurationWidget->functionLinearRadioButton->isChecked()) {
		newBlendingFunction = DeformSurfel::LINEAR;
	}
	else if (configurationWidget->functionQuadraticRadioButton->isChecked()) {
		newBlendingFunction = DeformSurfel::QUADRATIC;
	}

	if ( newBlendingFunction != currentBlendingFunction ) {

		currentBlendingFunction = newBlendingFunction;

		for (i = 0; i < numDeformSurfels; i++) {
			deformSurfels->at(i).computeDeformationScales (currentBlendingFunction);
		}

	}
}


void DeformerTool::splitSurfels (float minAxisRatio) {
	
	float           minAxisRatioSquared;
	unsigned int    i,
		            numDeformSurfelsFixed;
	QRgb			*color;
	
	
	minAxisRatioSquared = minAxisRatio * minAxisRatio;

	if (configurationWidget->ColorSplitSurfelsCheckBox->isChecked()) {
		color = new QRgb;
		*color = QColor(configurationWidget->RSpinBox->value(), configurationWidget->GSpinBox->value(), configurationWidget->BSpinBox->value()).rgb();
	}
	else {
		color = NULL;
	}
	// this will recursively split all surfels until the ratio of the squared axis length is below minAxisRatioSquared
	// right now this does not consider the actual size of the surfel only the ratio of the axes lengths!!
	
	// numDeformSurfels will grow, but we only want to consider original surfels
	numDeformSurfelsFixed = numDeformSurfels;
	for (i = 0; i < numDeformSurfelsFixed; i++) {		
		deformSurfels->at(i).splitSurfel (deformSurfels, numDeformSurfels, i, minAxisRatioSquared, 
			                              objectSurfels, currentBlendingFunction, scaleGrid, color);
	}

	if (color) {
		delete color;
	}
	
}




bool DeformerTool::checkToolAvailability() const {
	return RendererManager::getInstance()->getMainRenderer()->isInteractive() &&
		   (Scene::getInstance()->getNofObjects() > 0);
}


void DeformerTool::undoDeformation() {

	for (unsigned int i = 0; i < numDeformSurfels; i++) {			
		deformSurfels->at(i).undoDeformation();
	}

	Scene::getInstance()->emitObjectModifiedSignal (true);
	Scene::getInstance()->emitRenderFrameSignal ();
}

// *************
// private slots
// *************
void DeformerTool::handleToolSelected (int id) {

	currentTool = id;

	if (id == moveID) {
		toolButton->setIconSet (moveIconSet);
	}    
	else if (id == rotateID) {
		toolButton->setIconSet (rotateIconSet);
	}
}


void DeformerTool::handleActiveModelChanged() {
	if (!this->initDeformation()) {
		ToolManager::getInstance()->switchToNavigationTool();
	}
}


void DeformerTool::handleSceneMoved () {

	orthoMapper->reset();
	if (!isGlobalTransformationActive()) {
		currentPositionValid = orthoMapper->drawCursor (screenPosition, 0);
	}
}


void DeformerTool::handleViewPortSizeChanged () {
	orthoMapper->reset();
}


void DeformerTool::handlePreviewRendererToggled (bool isOn) {
	usePreviewRenderer = isOn;
}


void DeformerTool::showRotationAxes() {
	Scene::getInstance()->emitObjectModifiedSignal ();
	Scene::getInstance()->emitRenderFrameSignal ();
}


void DeformerTool::flagSelection() {
	Scene *scene = Scene::getInstance();

	QVector<SurfelInterface> *sel3 = CoreTools::getInstance()->getSelectionTool()->getSelection (SurfelInterface::SELECTED3);
	unsigned int cnt = sel3->count();
	for(unsigned int i=0; i<cnt; i++) {
		scene->setSurfelFlagOn(sel3->at(i), SurfelInterface::FEATURE, true, false);
	}

}

void DeformerTool::handleDynamicResamplingToggled (bool isOn) {
	useDynamicResampling = isOn;
	
	if (zombiePositions) {
		delete[] zombiePositions;
		zombiePositions = NULL;
		numZombies = 0;
		delete[] zombieColors;
		zombieColors = NULL;
	}

	if (useDynamicResampling) {
		numZombies = numDeformSurfels;
		if (numZombies > 0)
		zombiePositions = new Vector3D[numZombies];
		zombieColors = new QRgb[numZombies];
		for (unsigned int i=0; i<numZombies; i++) {
			zombiePositions[i] = deformSurfels->at(i).getSurfel()->getPosition();
			zombieColors[i] = deformSurfels->at(i).getSurfel()->getDiffuseColor();
		}
	}
}


void DeformerTool::handleInterpolation() {

	KdTree							*searchZombies;
	unsigned int					i, j, nOfFoundNeighbors;
	QColor							*diffColors,
									zombieDiffColor,
									newColor;
	float							r, g, b;
	SurfelInterface					*surfel;
	float							sqrSigma, maxSqrDist;
	float							gaussWeight, sumWeight;
	Scene							*scene;

	if (numZombies == 0) {
		return;
	}	
	
	scene = Scene::getInstance();
	diffColors = new QColor[numZombies];

	for (i = 0; i < numZombies; i++) {
		diffColors[i] = QColor(zombieColors[i]);
	}

	searchZombies = new KdTree(zombiePositions, numZombies, 10);
	searchZombies->setNOfNeighbours(configurationWidget->fairingneighborsSpinBox->value());
	
	for (i = 0; i < numDeformSurfels; i++) {

		r = g = b = 0;
		sumWeight = 0;

		surfel = deformSurfels->at(i).getSurfel();

		searchZombies->queryPosition(surfel->getPosition());
		nOfFoundNeighbors = searchZombies->getNOfFoundNeighbours();
		ASSERT(nOfFoundNeighbors);
		maxSqrDist = searchZombies->getSquaredDistance(nOfFoundNeighbors - 1); 
		sqrSigma = -maxSqrDist / log(0.01);		

		for (j = 0; j < nOfFoundNeighbors; j++) {
			gaussWeight = exp(-searchZombies->getSquaredDistance(j) / sqrSigma);
			sumWeight += gaussWeight;
			newColor = diffColors[searchZombies->getNeighbourPositionIndex(j)];
			r	+= gaussWeight * newColor.red();
			g	+= gaussWeight * newColor.green();
			b	+= gaussWeight * newColor.blue();
		}

		if (sumWeight > 0.0f) {
			r /= sumWeight;
			g /= sumWeight;
			b /= sumWeight;
		}

		scene->setSurfelDiffuseColor(surfel, QColor(r,g,b).rgb(), false);
	}

	delete searchZombies;
	delete[] diffColors;

	scene->emitObjectModifiedSignal ();
	scene->emitRenderFrameSignal ();

}

void DeformerTool::handleShowZombiesChanged() {
		
		Scene *scene = Scene::getInstance();
		SurfelInterface *surfel;
		unsigned int i;

		if (numZombies == 0) {
			return;
		}

		objectSurfels = scene->getActiveObject()->getSurfelCollection();

		if (configurationWidget->showZombiesCheckBox->isChecked()) {
			scene->setAllSurfelFlagsOn(SurfelInterface::COVERED, false, false);

			for (surfel = objectSurfels->getFirstSurfel(); surfel; surfel = objectSurfels->getNextSurfel()) {
				if (!surfel->isFlagOn(SurfelInterface::SELECTED1)) {
					scene->setSurfelFlagOn(surfel, SurfelInterface::INVISIBLE, true, false);
				}
			}
			
			for (i = 0; i < numZombies; i++) {
				surfel = objectSurfels->addSurfel(deformSurfels->at(i).getSurfel(), false);
				scene->setSurfelPosition(surfel, zombiePositions[i], false);
				scene->setSurfelDiffuseColor(surfel, zombieColors[i], false);
				scene->setSurfelFlagOn(surfel, SurfelInterface::COVERED, true, false);
				scene->setSurfelFlagOn(surfel, SurfelInterface::INVISIBLE, false, false);
			}
			
		}
		else {

			for (surfel = objectSurfels->getFirstSurfel(); surfel; ) {
				if (surfel->isFlagOn(SurfelInterface::COVERED)) {
					objectSurfels->removeCurrentSurfel();
					surfel = objectSurfels->getCurrentSurfel();
				}
				else {
					scene->setSurfelFlagOn(surfel, SurfelInterface::INVISIBLE, false, false);
					surfel = objectSurfels->getNextSurfel();
				}
			}
		}

		scene->emitObjectModifiedSignal ();
		scene->emitRenderFrameSignal ();
}


// ****************
// access functions
// ****************

ToolInterface *createTool() {
	return new DeformerTool();
}

const char *getToolName() {
	return "DeformerTool";
}





	

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
