// Title:   DeformerTool.h
// Created: Thu Sep 25 15:07:31 2003
// Authors: Richard Keiser, Mark Pauly
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __DEFORMERTOOL_H_
#define __DEFORMERTOOL_H_

#include <qwidget.h>
#include <qtoolbutton.h>
#include <qbuttongroup.h>
#include <qpoint.h>
#include <qpointarray.h>
#include <qcursor.h>
#include <qobject.h>
#include <qimage.h>
#include <qpopmenu.h>
#include <qcombobox.h>
#include <qfiledialog.h>

#include "../../../../DataTypes/src/MyDataTypes.h"
#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Utilities/src/Matrix.h"

#include "../../../../Core/RendererManager/src/RendererManager.h"
#include "../../../../Core/OrthoMapper/src/OrthoMapper.h"
#include "../../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../../Core/Interfaces/src/FrameBufferInterface.h"
#include "../../../../Core/Interfaces/src/ToolInterface.h"

#include "../UI/DeformerConfigurationWidget.h"
#include "DeformSurfel.h"
#include "ScaleGrid.h"

/**
 * This tool allows to perform free form deformation of a model.
 * The rigid region of the model needs to be selected with <code>SurfelInterface::SELECTION1</code>,
 * the maximal deformable region needs to be selected with <code>SurfelInterface::SELECTION2</code>,
 * while the remaining part of the deformalbe region is not selected.
 * 
 * @author Mark Pauly
 * @version 2.0
 */
class DeformerTool : public ToolInterface {

	Q_OBJECT

public:

	/**
	 * Creates and initializes this tool
	 */
	TOOL_API DeformerTool();
	/**
	 * Destructor
	 */
	TOOL_API virtual ~DeformerTool();

	// ***********************
	// interface ToolInterface
	// ***********************

	TOOL_API QToolButton *addButtonToToolBar (QToolBar *toolBar);	
	TOOL_API QCursor *getToolCursor() const;
	TOOL_API QToolButton *getToolButton();
	TOOL_API QWidget *getToolConfigurationWidget() const;
	TOOL_API bool isAvailable() const;

	TOOL_API void handleMousePressEvent   (QMouseEvent *me);
	TOOL_API void handleMouseReleaseEvent (QMouseEvent *me);
	TOOL_API void handleMouseMoveEvent    (QMouseEvent *me);
	TOOL_API void handleKeyReleaseEvent   (QKeyEvent   *ke);
	TOOL_API void handleKeyPressEvent	  (QKeyEvent   *ke);
	TOOL_API void selectTool();
	TOOL_API void deselectTool();

public slots:

	/**
	 * This method is called whenever the <code>RendererInterface</code> has repainted its
	 * renderer widget. The current tool may then redraw its contents with a <code>QPainter</code> into
	 * the <code>QWidget</code>.
	 *
	 * <b>Note:</b> This signal is <em>only</em> sent to this <code>ToolInterface</code> when
	 *              the tool is <em>selected</em>.
	 *
	 * @see RendererInterface#getRendererWidget
	 */
	void handleRendererWidgetRepainted();
	/**
	 * performs a relaxation of the surfels of the deformable region
	 */
	void handleFairing();


private:

	QPoint                      lastMouseDragPosition; // the last position of the mouse before the current mouse drag event
	QCursor                     *toolCursor;
	QCursor                     *standardCursor,
								*transformCursor;
	QIconSet                    moveIconSet,
		                        rotateIconSet;
	QToolButton                 *toolButton;	
	QPopupMenu                  deformerMenu;

	int                         moveID,
	                            rotateID,
						        currentTool;
	DeformerConfigurationWidget *configurationWidget;
	bool                        usePreviewRenderer;
	bool						useDynamicResampling;


	bool                     currentPositionValid;    // used for the orthomapper
	OrthoMapper              *orthoMapper;
	Vector2D                 screenPosition;          // mouse position in screen coordinates
	
	RendererInterface		 *mainRenderer;
	
	Vector3D                 *zeroRegionPositions,   // position vectors needed to build search structures
		                     *oneRegionPositions;


	QArray<DeformSurfel>     *deformSurfels;	      // array of all surfels that are in the deformed region
	unsigned int             numDeformSurfels;        // number of all surfels in the deformed region
	SurfelCollection         *objectSurfels;          // all surfels of deformed object


	NeighbourHood            *zeroRegion,             // search structures for zero-, one-region and the whole model
							 *oneRegion,
							 *origRegion;

	ScaleGrid				 *scaleGrid;


	Vector3D                 deformMoveVector;       // vector in which translation deformation (move) is performed
	float                    deformMoveScale;        // scale of translation deformation (move)
	Vector3D				 totalTranslation;				// the total deformation vector since start testing
	Vector3D				 *zombiePositions;		// array of zombies for interpolation during dynamic resampling
	QRgb					 *zombieColors;
	unsigned int			 numZombies;			// number of zombies


	unsigned int             numScaleNeighbors;      // number of nearest neighbors used to compute scale functions
	DeformSurfel::BlendingFunction currentBlendingFunction;  // blending function used to combine zero- and one-distance


	// rotation deformation data
	QPointArray	             frameShape;			// shape description of local frame
	QPointArray              rotationPieShape;      // shape description of rotation pie
	MyDataTypes::TransformationMatrix16f axisRotationMatrix,
		                                 axisScaleTranslationMatrix,
										 axisTransformationMatrix;	
	float				     rotationAngle;
	bool					 bIsSetRotationAxis;

	// called when the transformation is changed (scene transformation <=> local transformation)
	void setGlobalTransformationActive(bool isOn);
	
	// initializes deformation data
	// returns true if the initialization was successful, false otherwise
	bool initDeformation ();

	// recomputes the rotation axis 
	void updateRotationAxis();

	// renders the rotation axis
	void renderRotationAxis();

	// computes the deformation
	void deformModel (DeformParams &deformParams);
	
	// resamples the deformable surfles by splitting
	void splitSurfels (float minAxisRatio);
	
	// computes the deformation scale factors
	// only computated if blending function has changed or forceRecomputation is true
	void computeDeformationScales ();

	bool checkToolAvailability() const;

	void undoDeformation();
	
	DeformParams computeDeformParams(QMouseEvent *me);


private slots:
	void handleToolSelected (int id);
	void handleActiveModelChanged();
	void handleSceneMoved();
	void handleViewPortSizeChanged ();
	void handlePreviewRendererToggled (bool isOn);
	void showRotationAxes();
	void flagSelection();
	void handleDynamicResamplingToggled (bool isOn);
	void handleInterpolation();
	void handleShowZombiesChanged();

};

// ****************
// access functions
// ****************

// extern "C" is needed so those functions can be accessed
// with GetProcAddress() (WIN32) and dlsym() (UNIX) respective
extern "C" {
	TOOL_API ToolInterface *createTool();
	TOOL_API const char *getToolName();
}

#endif  // __DEFORMERTOOL_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
