// Title:   ScaleGrid.cpp
// Created: Thu Sep 25 15:07:28 2003
// Authors: Richard Keiser, Mark Pauly
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//


#include "ScaleGrid.h"
#include <math.h>



ScaleGrid::ScaleGrid (Vector3D *positions,
					  unsigned int nofPoints,
					  unsigned int nofGridPoints,
					  NeighbourHood *zeroRegion,
					  NeighbourHood *oneRegion) {

	unsigned int i,
		         x,
				 y,
				 z;
	float        volume;		// volume of bounding box
	float        avgSide;		// side length of same volume cube
	float        maxSide;		// maximum length of bounding box edge
	float        minNormDist;   // minimum normalized distance in grid

	Vector3D     point;

	//compute boundingbox
	bBox[0] = bBox[1] = positions[0][0];
	bBox[2] = bBox[3] = positions[0][1];
	bBox[4] = bBox[5] = positions[0][2];

	for (i = 1; i < nofPoints; i++) {
		if (bBox[0] > positions[i][0]) {	bBox[0] =  positions[i][0]; }
		if (bBox[1] < positions[i][0]) {	bBox[1] =  positions[i][0]; }
		if (bBox[2] > positions[i][1]) {	bBox[2] =  positions[i][1]; }
		if (bBox[3] < positions[i][1]) {	bBox[3] =  positions[i][1]; }
		if (bBox[4] > positions[i][2]) {	bBox[4] =  positions[i][2]; }
		if (bBox[5] < positions[i][2]) {	bBox[5] =  positions[i][2]; }
	}

	maxSide = bBox[1] - bBox[0];
	if (bBox[3] - bBox[2] > maxSide) { maxSide = bBox[3] - bBox[2]; }
	if (bBox[5] - bBox[4] > maxSide) { maxSide = bBox[5] - bBox[4]; }

	// enlarge boundingBox
	bBox[0] -= maxSide / (float) nofGridPoints;
	bBox[1] += maxSide / (float) nofGridPoints;
	bBox[2] -= maxSide / (float) nofGridPoints;
	bBox[3] += maxSide / (float) nofGridPoints;
	bBox[4] -= maxSide / (float) nofGridPoints;
	bBox[5] += maxSide / (float) nofGridPoints;

	volume = (bBox[1] - bBox[0]) * (bBox[3] - bBox[2]) * (bBox[5] - bBox[4]);
	avgSide = pow (volume, 0.33333333f);

	// compute grid size
	sizeX = (float) nofGridPoints * (bBox[1] - bBox[0]) / avgSide;
	sizeY = (float) nofGridPoints * (bBox[3] - bBox[2]) / avgSide;
	sizeZ = (float) nofGridPoints * (bBox[5] - bBox[4]) / avgSide;

	// at least two grid points per dimension ( -> trilinear interpolation)
	if (sizeX < 2) { sizeX = 2; }
	if (sizeY < 2) { sizeY = 2; }
	if (sizeZ < 2) { sizeZ = 2; }

	cellSizes[0] = (bBox[1] - bBox[0]) / (float) (sizeX - 1);
	cellSizes[1] = (bBox[3] - bBox[2]) / (float) (sizeY - 1);
	cellSizes[2] = (bBox[5] - bBox[4]) / (float) (sizeZ - 1);

	
	grid = new float [sizeX * sizeY * sizeZ];

	
	minNormDist = FLT_MAX;

	for (x = 0; x < sizeX; x++) {
		for (y = 0; y < sizeY; y++) {
			for (z = 0; z < sizeZ; z++) {
				point[0] = bBox[0] + x * cellSizes[0];
				point[1] = bBox[2] + y * cellSizes[1];
				point[2] = bBox[4] + z * cellSizes[2];

				grid[x*sizeY*sizeZ + y*sizeZ + z] = this->computeNormalizedDistance (zeroRegion, oneRegion, point);

				if (grid[x*sizeY*sizeZ + y*sizeZ + z] < minNormDist) {
					minNormDist = grid[x*sizeY*sizeZ + y*sizeZ + z];
				}
			}
		}
	}


	for (x = 0; x < sizeX; x++) {
		for (y = 0; y < sizeY; y++) {
			for (z = 0; z < sizeZ; z++) {
				grid[x*sizeY*sizeZ + y*sizeZ + z] -= minNormDist * 4;
				if (grid[x*sizeY*sizeZ + y*sizeZ + z] < 0) {
					grid[x*sizeY*sizeZ + y*sizeZ + z] = 0;
				}
			}
		}
	}





}



// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
